import * as THREE from "../build/three.module.js"

class App {
    constructor() {
        this._initialize();
    }

    _initialize() {
        const domWebGL = document.createElement('div');
        document.body.appendChild(domWebGL);
        this._domWebGL = domWebGL;

        const renderer = new THREE.WebGLRenderer({ antialias: true });
        renderer.outputEncoding = THREE.sRGBEncoding;
        renderer.setPixelRatio(window.devicePixelRatio);
        renderer.shadowMap.enabled =  true;
        domWebGL.appendChild(renderer.domElement);
        this._renderer = renderer;

        const scene = new THREE.Scene();
        scene.background = new THREE.Color(0x444488);
        this._scene = scene;
        
        this._setupCamera();
        this._setupLight();
        this._setupModel();
        //this._setupInteractions();
       
        //this._setupGUI();
        //this._setupControls();
        //this._setupHelper();
        
        window.onresize = this.resize.bind(this);

        this.resize();
        this.render()
    }

    _setupInteractions() {
        //.
    }

    _setupPostprocessing() {
        //.
    }

    _setupModel(font) {
        const geometryBackground = new THREE.PlaneGeometry(100, 100);
        const materialBackground = new THREE.MeshPhongMaterial({color: 0x000066});

        const background = new THREE.Mesh(geometryBackground, materialBackground);
        background.receiveShadow = true;
        background.position.set(0,0,-1);
        this._scene.add(background);

        const geometryCylinder = new THREE.CylinderGeometry(0.5,0.5,1,32); 
        const materialCylinder = new THREE.MeshPhongMaterial({color: 0xff0000 });
        const mesh = new THREE.Mesh(geometryCylinder, materialCylinder);
        mesh.castShadow = true;
        mesh.receiveShadow = true;
        this._scene.add(mesh);

        this._mesh = mesh;
    }

    _addLabel(font, name, location) {
        const textGeo = new THREE.TextGeometry(name, {
            font: font,
            size: 20,
            height: 1,
            curveSegments: 1
        });

        const textMaterial = new THREE.MeshBasicMaterial({ color: 0xffffff });
        const textMesh = new THREE.Mesh(textGeo, textMaterial);
        textMesh.position.copy(location);

        this._scene.add(textMesh);
    }

    _setupHelper() {
        //.
    }

    _setupGUI() {
        const stats = new Stats();
        this._domWebGL.appendChild(stats.dom);
        this._stats = stats;
    }

    _setupCamera() {    
        const AMOUNT = 3;
        const ASPECT_RATIO = window.innerWidth / window.innerHeight;
        const WIDTH = (window.innerWidth / AMOUNT) * window.devicePixelRatio;
        const HEIGHT = (window.innerHeight / AMOUNT) * window.devicePixelRatio;
        const cameras = [];

        for(let y=0; y<AMOUNT; y++) {
            for(let x=0; x<AMOUNT; x++) {
                const subcamera = new THREE.PerspectiveCamera(40, ASPECT_RATIO, 0.1, 10);
                subcamera.viewport = new THREE.Vector4(
                    Math.floor(x*WIDTH), 
                    Math.floor(y*HEIGHT), 
                    Math.ceil(WIDTH),
                    Math.ceil(HEIGHT)
                );
                subcamera.position.set(
                    (x/AMOUNT)-0.5,
                    0.5-(y/AMOUNT),
                    1.5
                );

                subcamera.position.multiplyScalar(2 + Math.pow(x/AMOUNT + y/AMOUNT, 2));
                subcamera.lookAt(0,0,0);
                subcamera.updateMatrixWorld();
                
                cameras.push(subcamera);
            }
        }

        const camera = new THREE.ArrayCamera(cameras);
        camera.position.z = 3;
    
        this._camera = camera;
        this._AMOUNT = AMOUNT;

    }

    _setupControls() {
        const controls = new OrbitControls(this._camera, this._renderer.domElement);
        controls.minDistance = 200;
        controls.maxDistance = 2000;
    }

    _setupLight() {
        this._scene.add(new THREE.AmbientLight(0x222244));
        const light = new THREE.DirectionalLight();
        light.position.set(0.5, 0.5, 1);
        light.castShadow = true;
        light.shadow.camera.zoom = 4;
        this._scene.add(light);
    }

    update() {
        this._mesh.rotation.x += 0.005;
        this._mesh.rotation.z += 0.01;
    }

    render() {
        requestAnimationFrame(this.render.bind(this));
        
        this.update();
        this._renderer.render(this._scene, this._camera);
    }

    resize() {
        const AMOUNT = this._AMOUNT;

        const aspect = window.innerWidth / window.innerHeight;
        const width = (window.innerWidth / AMOUNT) * window.devicePixelRatio;
        const height = (window.innerHeight / AMOUNT) * window.devicePixelRatio;

        this._camera.aspect = aspect;
        this._camera.updateProjectionMatrix();

        for(let y=0; y<AMOUNT; y++) {
            for(let x=0; x<AMOUNT; x++) {
                const subcamera = this._camera.cameras[AMOUNT*y+x];
                subcamera.viewport.set(
                    Math.floor(x*width),
                    Math.floor(y*height),
                    Math.ceil(width),
                    Math.ceil(height)
                );
                subcamera.aspect = aspect;
                subcamera.updateProjectionMatrix();
            }
        }

        this._renderer.setSize(window.innerWidth, window.innerHeight);
    }
}

window.onload = function () {
    new App();
}